﻿using Fclp;
using Nito.AsyncEx;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading;
using System.Threading.Tasks;
using Windows.Devices.Bluetooth;
using Windows.Devices.Bluetooth.GenericAttributeProfile;
using Windows.Devices.Enumeration;

namespace gatttool
{
    static class Program
    {
        public static readonly ManualResetEventSlim WaitHandle = new ManualResetEventSlim(false);
        public static readonly List<string> DeviceIds = new List<string>();
        private const string Hex = "[0-9a-fA-F]";
        public static readonly string ByteInHex = Hex + Hex;
        public static readonly string Mac = string.Join(@"\:", ByteInHex, ByteInHex, ByteInHex, ByteInHex, ByteInHex, ByteInHex);
        public static readonly Regex BtIdRegex = new Regex(@"^BluetoothLE\#BluetoothLE" + Mac + "-(?<Mac>" + Mac +")$", RegexOptions.Compiled);
        public static readonly Regex MacRegex = new Regex(@"^" + Mac + "$", RegexOptions.Compiled);
        public static readonly Regex HandleRegex = new Regex(@"^0x" + Hex + "+$", RegexOptions.Compiled);
        public static readonly Regex HexRegex = new Regex(@"^(" + ByteInHex + ")+$", RegexOptions.Compiled);
        private static Parameters parameters;

        static int Main(string[] args)
        {
            if (!ParseCommandLine(args))
            {
                Console.WriteLine("Press Enter to exit");
                Console.ReadLine();
                return 1;
            }

            if (!AsyncContext.Run(MainAsync))
            {
                Console.WriteLine("Press Enter to exit");
                Console.ReadLine();
                return 1;
            }

            Console.WriteLine("Press Enter to exit");
            Console.ReadLine();

            return 0;
        }

        static bool ParseCommandLine(string[] args)
        {
            var p = new FluentCommandLineParser<Parameters>();
            p.Setup(arg => arg.DeviceMac).As('b', "device").Required();
            p.Setup(arg => arg.Read).As("char-read").SetDefault(false);
            p.Setup(arg => arg.Write).As("char-write-req").SetDefault(false);
            p.Setup(arg => arg.Characteristics).As("characteristics").SetDefault(false);
            p.Setup(arg => arg.CharacteristicHandle).As('a', "handle");
            p.Setup(arg => arg.WriteValue).As('n', "value").SetDefault(string.Empty);

            var result = p.Parse(args);

            if (result.Errors.Any())
            {
                Console.WriteLine(result.ErrorText);
                return false;
            }

            parameters = p.Object;
            if (!MacRegex.IsMatch(parameters.DeviceMac))
            {
                Console.WriteLine("Invalid Mac address provided");
                return false;
            }

            if (parameters.Read || parameters.Write)
            {
                if (string.IsNullOrEmpty(parameters.CharacteristicHandle))
                {
                    Console.WriteLine("handle must be provided when reading or writing");
                    return false;
                }
                if (!HandleRegex.IsMatch(parameters.CharacteristicHandle))
                {
                    Console.WriteLine("Invalid characteristic handle provided");
                    return false;
                }

            }

            if (!parameters.Write && !parameters.Read && !parameters.Characteristics)
            {
                Console.WriteLine("Must either set read, write or characteristics mode!");
                return false;
            }

            if (parameters.Write && parameters.Read)
            {
                Console.WriteLine("Can't set both read and write mode!");
                return false;
            }

            if (parameters.Write && string.IsNullOrWhiteSpace(parameters.WriteValue))
            {
                Console.WriteLine("When in write mode write value must be provided!");
                return false;
            }

            if (parameters.Write && !HexRegex.IsMatch(parameters.WriteValue))
            {
                Console.WriteLine("Write value is invalid must be Hex!");
                return false;
            }

            return true;
        }

        static async Task<bool> MainAsync()
        {
            try
            {
                var deviceWatcher = DeviceInformation.CreateWatcher(BluetoothLEDevice.GetDeviceSelector());
                deviceWatcher.Added += DeviceWatcher_Added;
                deviceWatcher.EnumerationCompleted += DeviceWatcher_EnumerationCompleted;
                Trace.WriteLine("Enumerating devices");
                deviceWatcher.Start();
                WaitHandle.Wait();
                Trace.WriteLine("Done");
                deviceWatcher.Stop();

                if (DeviceIds.Count == 0)
                {
                    Console.WriteLine("Failed to find device with correct Mac address");
                    return false;
                }
                if (DeviceIds.Count > 1)
                {
                    Console.WriteLine("Found multiple devices with Mac address");
                    return false;
                }
                var device = await BluetoothLEDevice.FromIdAsync(DeviceIds.First());

                //Look up uuid
                if (parameters.Characteristics)
                {
                    var characteristics = device.GattServices.
                                                 SelectMany(s => s.GetAllCharacteristics());
                    foreach (var characteristic in characteristics)
                    {
                        //We are guessing that char value handle = handle + 1, this might be incorrect.
                        Console.WriteLine($"handle = 0x{characteristic.AttributeHandle:x4}, char properties = 0x{((int)characteristic.CharacteristicProperties):x2}, char value handle = 0x{characteristic.AttributeHandle + 1:x4}, uuid = {characteristic.Uuid}");
                    }
                    return true;
                }
                else
                {
                    var characteristicValueHandle = int.Parse(parameters.CharacteristicHandle.Substring(2), NumberStyles.HexNumber);
                    //We are guessing that handle = char value handle - 1, this might be incorrect.
                    var characteristicHandle = characteristicValueHandle - 1;
                    var characteristic = device.GattServices.
                                                SelectMany(s => s.GetAllCharacteristics()).
                                                FirstOrDefault(c => c.AttributeHandle == characteristicHandle);

                    if (characteristic == null)
                    {
                        Console.WriteLine("Failed to find characteristic");
                        return false;
                    }
                    if (parameters.Read)
                    {
                        return await ReadCharacteristic(characteristic);
                    }
                    else
                    {
                        //Parse write value
                        var data = StringToByteArray(parameters.WriteValue);
                        return await WriteCharacteristic(characteristic, data);
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
                return false;
            }
        }

        private static void DeviceWatcher_EnumerationCompleted(DeviceWatcher sender, object args)
        {
            WaitHandle.Set();
        }

        //private static void DeviceWatcher_Updated(DeviceWatcher sender, DeviceInformationUpdate args)
        //{
        //    Console.WriteLine(System.Threading.Thread.CurrentThread.ManagedThreadId);
        //}

        private static void DeviceWatcher_Added(DeviceWatcher sender, DeviceInformation args)
        {
            var match = BtIdRegex.Match(args.Id);
            if (match.Success)
            {
                var mac = match.Groups["Mac"].Value;
                if (mac.Equals(parameters.DeviceMac, StringComparison.OrdinalIgnoreCase))
                    DeviceIds.Add(args.Id);
            }
        }

        public static byte[] StringToByteArray(string hex)
        {
            int numberChars = hex.Length;
            byte[] bytes = new byte[numberChars / 2];
            for (int i = 0; i < numberChars; i += 2)
                bytes[i / 2] = Convert.ToByte(hex.Substring(i, 2), 16);
            return bytes;
        }

        private static async Task<bool> ReadCharacteristic(GattCharacteristic characteristic)
        {
            var value = await characteristic.ReadValueAsync(BluetoothCacheMode.Uncached);
            if (value.Status == GattCommunicationStatus.Success)
            {
                var reader = Windows.Storage.Streams.DataReader.FromBuffer(value.Value);
                var buffer = new byte[value.Value.Length];
                reader.ReadBytes(buffer);
                Console.WriteLine($"Characteristic value/descriptor: {BitConverter.ToString(buffer).Replace("-", " ")}");
                return true;
            }

            Console.WriteLine("connect error: Transport endpoint is not connected (107)");
            return false;
        }
        private static async Task<bool> WriteCharacteristic(GattCharacteristic characteristic, byte[] data)
        {
            var writer = new Windows.Storage.Streams.DataWriter();
            writer.WriteBytes(data);
            var result = await characteristic.WriteValueAsync(writer.DetachBuffer(), GattWriteOption.WriteWithResponse);
            if (result == GattCommunicationStatus.Success)
            {
                Console.WriteLine("Characteristic value was written successfully");
                return true;
            }

            Console.WriteLine("connect error: Transport endpoint is not connected (107)");
            return false;
        }
    }
}
